# The Flask HTTP Digest Authentication Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2022/11/3

#  Copyright (c) 2022 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The algorithm.

"""
from __future__ import annotations

from hashlib import md5
from typing import Optional, Literal


def make_password_hash(realm: str, username: str, password: str) -> str:
    """Calculates the password hash for the HTTP digest authentication.
    Use this function to set the password for the user.

    :Example:

    ::

        user.password = make_password_hash(realm, username, password)

    :param realm: The realm.
    :param username: The username.
    :param password: The cleartext password.
    :return: The password hash for the HTTP digest authentication.
    """
    return md5(f"{username}:{realm}:{password}".encode("utf8")).hexdigest()


def calc_response(
        method: str, uri: str, password_hash: str,
        nonce: str, qop: Optional[Literal["auth", "auth-int"]] = None,
        algorithm: Optional[Literal["MD5", "MD5-sess"]] = "MD5-sess",
        cnonce: Optional[str] = None, nc: Optional[str] = None,
        body: Optional[bytes] = None) -> str:
    """Calculates the response value of the HTTP digest authentication.

    :param method: The request method.
    :param uri: The request URI.
    :param password_hash: The password hash for the HTTP digest authentication.
    :param nonce: The nonce.
    :param qop: The quality of protection, either ``auth`` or ``auth-int``.
    :param algorithm: The algorithm, either ``MD5`` or ``MD5-sess``.
    :param cnonce: The client nonce, which must exists when qop exists or
        algorithm is ``MD5-sess``.
    :param nc: The request counter, which must exists when qop exists.
    :param body: The request body, which must exists when qop is ``auth-int``.
    :return: The response value.
    :raise AssertionError: When cnonce is missing with algorithm is
        ``MD5-sess``, when body is missing with qop is ``auth-int``, or when
        cnonce or nc is missing with qop exits.
    """

    def calc_ha1() -> str:
        """Calculates and returns the first hash.

        :return: The first hash.
        :raise AssertionError: When cnonce is missing with
            algorithm is ``MD5-sess``.
        """
        if algorithm == "MD5-sess":
            assert cnonce is not None,\
                f"Missing \"cnonce\" with algorithm=\"{algorithm}\""
            return md5(f"{password_hash}:{nonce}:{cnonce}".encode("utf8")) \
                .hexdigest()
        # algorithm is None or algorithm == "MD5"
        return password_hash

    def calc_ha2() -> str:
        """Calculates the second hash.

        :return: The second hash.
        :raise AssertionError: When body is missing with qop is ``auth-int``.
        """
        if qop == "auth-int":
            assert body is not None, f"Missing \"body\" with qop=\"{qop}\""
            return md5(
                f"{method}:{uri}:{md5(body).hexdigest()}".encode("utf8")) \
                .hexdigest()
        # qop is None or qop == "auth"
        return md5(f"{method}:{uri}".encode("utf8")).hexdigest()

    ha1: str = calc_ha1()
    ha2: str = calc_ha2()
    if qop == "auth" or qop == "auth-int":
        assert cnonce is not None, f"Missing \"cnonce\" with the qop=\"{qop}\""
        assert nc is not None, f"Missing \"nc\" with the qop=\"{qop}\""
        return md5(f"{ha1}:{nonce}:{nc}:{cnonce}:{qop}:{ha2}".encode("utf8"))\
            .hexdigest()
    # qop is None
    return md5(f"{ha1}:{nonce}:{ha2}".encode("utf8")).hexdigest()



# The Flask HTTP Digest Authentication Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2022/11/23

#  Copyright (c) 2022-2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The test case for the Flask-Login integration.

"""
import logging
import unittest
from secrets import token_urlsafe
from typing import Optional, Dict

import httpx
from flask import Flask, g, redirect, request
from werkzeug.datastructures import WWWAuthenticate

from flask_digest_auth import DigestAuth, make_password_hash
from testlib import REALM, USERNAME, PASSWORD, ADMIN_1_URI, ADMIN_2_URI, \
    LOGOUT_URI, make_authorization

SKIPPED_NO_FLASK_LOGIN: str = "Skipped without Flask-Login."
"""The message that a test is skipped when Flask-Login is not installed."""


class User:
    """A dummy user."""

    def __init__(self, username: str, password: str):
        """Constructs a dummy user.

        :param username: The username.
        :param password: The clear-text password.
        """
        self.username: str = username
        """The username."""
        self.password_hash: str = make_password_hash(REALM, username, password)
        """The password hash."""
        self.visits: int = 0
        """The number of visits."""
        self.is_active: bool = True
        """True if the account is active, or False otherwise."""
        self.is_anonymous: bool = False
        """True if the account is anonymous, or False otherwise."""

    def get_id(self) -> str:
        """Returns the username.
        This is required by Flask-Login.

        :return: The username.
        """
        return self.username

    @property
    def is_authenticated(self) -> bool:
        """Returns whether the user is authenticated.
        This is required by Flask-Login.
        This should return self.is_active.

        :return: True if the user is active, or False otherwise.
        """
        return self.is_active


class FlaskLoginTestCase(unittest.TestCase):
    """The test case with the Flask-Login integration."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        logging.getLogger("test_flask_login").addHandler(logging.NullHandler())
        self.app: Flask = Flask(__name__)
        self.app.config.from_mapping({
            "TESTING": True,
            "SECRET_KEY": token_urlsafe(32),
            "DIGEST_AUTH_REALM": REALM,
        })
        self.__client: httpx.Client = httpx.Client(
            transport=httpx.WSGITransport(app=self.app),
            base_url="https://testserver")
        """The testing client."""

        self.__has_flask_login: bool = True
        """Whether the Flask-Login package is installed."""
        try:
            import flask_login
        except ModuleNotFoundError:
            self.__has_flask_login = False
            return
        except ImportError:
            self.__has_flask_login = False
            return

        login_manager: flask_login.LoginManager = flask_login.LoginManager()
        login_manager.init_app(self.app)

        auth: DigestAuth = DigestAuth()
        auth.init_app(self.app)

        self.__user: User = User(USERNAME, PASSWORD)
        """The user account."""
        user_db: Dict[str, User] = {USERNAME: self.__user}

        @auth.register_get_password
        def get_password_hash(username: str) -> Optional[str]:
            """Returns the password hash of a user.

            :param username: The username.
            :return: The password hash, or None if the user does not exist.
            """
            return user_db[username].password_hash if username in user_db \
                else None

        @auth.register_on_login
        def on_login(user: User):
            """The callback when the user logs in.

            :param user: The logged-in user.
            :return: None.
            """
            user.visits = user.visits + 1

        @login_manager.user_loader
        def load_user(user_id: str) -> Optional[User]:
            """Loads a user.

            :param user_id: The username.
            :return: The user, or None if the user does not exist.
            """
            return user_db[user_id] if user_id in user_db else None

        @self.app.get(ADMIN_1_URI)
        @flask_login.login_required
        def admin_1() -> str:
            """The first administration section.

            :return: The response.
            """
            return f"Hello, {flask_login.current_user.get_id()}! #1"

        @self.app.get(ADMIN_2_URI)
        @flask_login.login_required
        def admin_2() -> str:
            """The second administration section.

            :return: The response.
            """
            return f"Hello, {flask_login.current_user.get_id()}! #2"

        @self.app.post(LOGOUT_URI)
        @flask_login.login_required
        def logout() -> redirect:
            """Logs out the user.

            :return: The response.
            """
            auth.logout()
            return redirect(request.form.get("next"))

    def test_auth(self) -> None:
        """Tests the authentication.

        :return: None.
        """
        if not self.__has_flask_login:
            self.skipTest(SKIPPED_NO_FLASK_LOGIN)

        response: httpx.Response

        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 401)
        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.text, f"Hello, {USERNAME}! #1")
        response = self.__client.get(ADMIN_2_URI)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.text, f"Hello, {USERNAME}! #2")
        self.assertEqual(self.__user.visits, 1)

    def test_stale_opaque(self) -> None:
        """Tests the stale and opaque value.

        :return: None.
        """
        if not self.__has_flask_login:
            self.skipTest(SKIPPED_NO_FLASK_LOGIN)

        response: httpx.Response
        www_authenticate: WWWAuthenticate
        auth_header: str

        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 401)
        www_authenticate = WWWAuthenticate.from_header(
            response.headers["WWW-Authenticate"])
        self.assertEqual(www_authenticate.type, "digest")
        self.assertIsNone(www_authenticate.get("stale"))
        opaque: str = www_authenticate.opaque

        with self.app.app_context():
            if hasattr(g, "_login_user"):
                delattr(g, "_login_user")
        www_authenticate.nonce = "bad"
        auth_header = make_authorization(
            www_authenticate, ADMIN_1_URI, USERNAME, PASSWORD)
        response = self.__client.get(ADMIN_1_URI,
                                     headers={"Authorization": auth_header})
        self.assertEqual(response.status_code, 401)
        www_authenticate = WWWAuthenticate.from_header(
            response.headers["WWW-Authenticate"])
        self.assertEqual(www_authenticate.get("stale"), "TRUE")
        self.assertEqual(www_authenticate.opaque, opaque)

        with self.app.app_context():
            if hasattr(g, "_login_user"):
                delattr(g, "_login_user")
        auth_header = make_authorization(
            www_authenticate, ADMIN_1_URI, USERNAME, PASSWORD + "2")
        response = self.__client.get(ADMIN_1_URI,
                                     headers={"Authorization": auth_header})
        self.assertEqual(response.status_code, 401)
        www_authenticate = WWWAuthenticate.from_header(
            response.headers["WWW-Authenticate"])
        self.assertEqual(www_authenticate.get("stale"), "FALSE")
        self.assertEqual(www_authenticate.opaque, opaque)

        with self.app.app_context():
            if hasattr(g, "_login_user"):
                delattr(g, "_login_user")
        auth_header = make_authorization(
            www_authenticate, ADMIN_1_URI, USERNAME, PASSWORD)
        response = self.__client.get(ADMIN_1_URI,
                                     headers={"Authorization": auth_header})
        self.assertEqual(response.status_code, 200)

    def test_logout(self) -> None:
        """Tests the logging out.

        :return: None.
        """
        if not self.__has_flask_login:
            self.skipTest(SKIPPED_NO_FLASK_LOGIN)

        response: httpx.Response

        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 401)

        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 200)

        response = self.__client.post(LOGOUT_URI, data={"next": ADMIN_1_URI})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], ADMIN_1_URI)

        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 401)

        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 401)

        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 200)

        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.__user.visits, 2)

    def test_disabled(self) -> None:
        """Tests the disabled user.

        :return: None.
        """
        if not self.__has_flask_login:
            self.skipTest(SKIPPED_NO_FLASK_LOGIN)

        response: httpx.Response

        self.__user.is_active = False
        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 401)
        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 401)

        self.__user.is_active = True
        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 200)
        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 200)

        self.__user.is_active = False
        response = self.__client.get(ADMIN_1_URI)
        self.assertEqual(response.status_code, 401)
        response = self.__client.get(ADMIN_1_URI,
                                     auth=httpx.DigestAuth(USERNAME, PASSWORD))
        self.assertEqual(response.status_code, 401)



# The Flask HTTP Digest Authentication Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2022/10/22

#  Copyright (c) 2022-2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The HTTP Digest Authentication.
See `RFC 2617`_ HTTP Authentication: Basic and Digest Access Authentication

.. _RFC 2617: https://www.rfc-editor.org/rfc/rfc2617
"""
from __future__ import annotations

import sys
from functools import wraps
from secrets import token_urlsafe, randbits
from typing import Any, Optional, Literal, Callable, List

from flask import g, request, Response, session, abort, Flask, Request, \
    current_app
from itsdangerous import URLSafeTimedSerializer, BadData
from werkzeug.datastructures import Authorization

from flask_digest_auth.algo import calc_response


class DigestAuth:
    """The HTTP digest authentication."""

    def __init__(self, realm: Optional[str] = None):
        """Constructs the HTTP digest authentication.

        :param realm: The realm.
        """
        self.__serializer: URLSafeTimedSerializer \
            = URLSafeTimedSerializer(token_urlsafe(32))
        """The serializer to generate and validate the nonce and opaque."""
        self.realm: str = "Login Required" if realm is None else realm
        """The realm.  Default is "Login Required"."""
        self.algorithm: Optional[Literal["MD5", "MD5-sess"]] = None
        """The algorithm, either None, ``MD5``, or ``MD5-sess``.  Default is
        None."""
        self.use_opaque: bool = True
        """Whether to use an opaque.  Default is True."""
        self.__domain: List[str] = []
        """A list of directories that this username and password applies to.
        Default is empty."""
        self.__qop: List[Literal["auth", "auth-int"]] = ["auth", "auth-int"]
        """A list of supported quality of protection supported, either
        ``qop``, ``auth-int``, both, or empty.  Default is both."""
        self.__get_password_hash: BasePasswordHashGetter \
            = BasePasswordHashGetter()
        """The callback to return the password hash."""
        self.__get_user: BaseUserGetter = BaseUserGetter()
        """The callback to return the user."""
        self.__on_login: BaseOnLogInCallback = BaseOnLogInCallback()
        """The callback to run when the user logs in."""

    def login_required(self, view) -> Callable:
        """The view decorator for the HTTP digest authentication.

        :Example:

        ::

            @app.get("/admin")
            @auth.login_required
            def admin():
                return f"Hello, {g.user.username}!"

        The logged-in user can be retrieved at ``g.user``.

        :param view: The view.
        :return: The login-protected view.
        """

        class NoLogInException(Exception):
            """The exception thrown when the user is not authorized."""

        def get_logged_in_user() -> Any:
            """Returns the currently logged-in user.

            :return: The currently logged-in user.
            :raise NoLogInException: When the user is not logged in.
            """
            if "user" not in session:
                raise NoLogInException
            user: Optional[Any] = self.__get_user(session["user"])
            if user is None:
                del session["user"]
                raise NoLogInException
            return user

        def auth_user(state: AuthState) -> Any:
            """Authenticates a user.

            :param state: The authentication state.
            :return: The user.
            :raise UnauthorizedException: When the authentication fails.
            """
            authorization: Authorization = request.authorization
            if authorization is None:
                raise UnauthorizedException
            if authorization.type != "digest":
                raise UnauthorizedException(
                    "Not an HTTP digest authorization")
            self.__authenticate(state)
            session["user"] = authorization.username
            return self.__get_user(authorization.username)

        @wraps(view)
        def login_required_view(*args, **kwargs) -> Any:
            """The login-protected view.

            :param args: The positional arguments of the view.
            :param kwargs: The keyword arguments of the view.
            :return: The response.
            """
            try:
                g.user = get_logged_in_user()
                return view(*args, **kwargs)
            except NoLogInException:
                pass

            state: AuthState = AuthState()
            try:
                g.user = auth_user(state)
                self.__on_login(g.user)
                return view(*args, **kwargs)
            except UnauthorizedException as e:
                if len(e.args) > 0:
                    current_app.logger.warning(e.args[0])
                response: Response = Response()
                response.status = 401
                response.headers["WWW-Authenticate"] \
                    = self.__make_response_header(state)
                abort(response)

        return login_required_view

    def __authenticate(self, state: AuthState) -> None:
        """Authenticate a user.

        :param state: The authorization state.
        :return: None.
        :raise UnauthorizedException: When the authentication failed.
        """
        if "digest_auth_logout" in session:
            del session["digest_auth_logout"]
            raise UnauthorizedException("Logging out")
        authorization: Authorization = request.authorization
        if self.use_opaque:
            if authorization.opaque is None:
                raise UnauthorizedException(
                    "Missing \"opaque\" in the Authorization header")
            try:
                self.__serializer.loads(
                    authorization.opaque, salt="opaque", max_age=1800)
            except BadData:
                raise UnauthorizedException("Invalid opaque")
            state.opaque = authorization.opaque
        password_hash: Optional[str] \
            = self.__get_password_hash(authorization.username)
        if password_hash is None:
            raise UnauthorizedException(
                f"No such user \"{authorization.username}\"")
        expected: str = calc_response(
            method=request.method, uri=authorization.uri,
            password_hash=password_hash, nonce=authorization.nonce,
            qop=authorization.qop,
            algorithm=authorization.get("algorithm"),
            cnonce=authorization.cnonce, nc=authorization.nc,
            body=request.data)
        if authorization.response != expected:
            state.stale = False
            raise UnauthorizedException("Incorrect response value")
        try:
            self.__serializer.loads(
                authorization.nonce,
                salt="nonce" if authorization.opaque is None
                else f"nonce-{authorization.opaque}")
        except BadData:
            state.stale = True
            raise UnauthorizedException("Invalid nonce")

    def __make_response_header(self, state: AuthState) -> str:
        """Composes and returns the ``WWW-Authenticate`` response header.

        :param state: The authorization state.
        :return: The ``WWW-Authenticate`` response header.
        """

        def get_opaque() -> Optional[str]:
            """Returns the opaque value.

            :return: The opaque value.
            """
            if not self.use_opaque:
                return None
            if state.opaque is not None:
                return state.opaque
            return self.__serializer.dumps(randbits(32), salt="opaque")

        opaque: Optional[str] = get_opaque()
        nonce: str = self.__serializer.dumps(
            randbits(32),
            salt="nonce" if opaque is None else f"nonce-{opaque}")

        header: str = f"Digest realm=\"{self.realm}\""
        if len(self.__domain) > 0:
            domain_list: str = ",".join(self.__domain)
            header += f", domain=\"{domain_list}\""
        header += f", nonce=\"{nonce}\""
        if opaque is not None:
            header += f", opaque=\"{opaque}\""
        if state.stale is not None:
            header += ", stale=TRUE" if state.stale else ", stale=FALSE"
        if self.algorithm is not None:
            header += f", algorithm=\"{self.algorithm}\""
        if len(self.__qop) > 0:
            qop_list: str = ",".join(self.__qop)
            header += f", qop=\"{qop_list}\""
        return header

    def register_get_password(self, func: Callable[[str], Optional[str]]) \
            -> None:
        """The decorator to register the callback to obtain the password hash.

        :Example:

        ::

            @auth.register_get_password
            def get_password_hash(username: str) -> Optional[str]:
                user = User.query.filter(User.username == username).first()
                return None if user is None else user.password

        :param func: The callback that given the username, returns the password
            hash, or None if the user does not exist.
        :return: None.
        """

        class PasswordHashGetter(BasePasswordHashGetter):
            """The base password hash getter."""

            @staticmethod
            def __call__(username: str) -> Optional[str]:
                """Returns the password hash of a user.

                :param username: The username.
                :return: The password hash, or None if the user does not exist.
                """
                return func(username)

        self.__get_password_hash = PasswordHashGetter()

    def register_get_user(self, func: Callable[[str], Optional[Any]]) -> None:
        """The decorator to register the callback to obtain the user.

        :Example:

        ::

            @auth.register_get_user
            def get_user(username: str) -> Optional[User]:
                return User.query.filter(User.username == username).first()

        :param func: The callback that given the username, returns the user,
            or None if the user does not exist.
        :return: None.
        """

        class UserGetter(BaseUserGetter):
            """The user getter."""

            @staticmethod
            def __call__(username: str) -> Optional[Any]:
                """Returns a user.

                :param username: The username.
                :return: The user, or None if the user does not exist.
                """
                return func(username)

        self.__get_user = UserGetter()

    def register_on_login(self, func: Callable[[Any], None]) -> None:
        """The decorator to register the callback to run when the user logs in.

        :Example:

        ::

            @auth.register_on_login
            def on_login(user: User) -> None:
                user.visits = user.visits + 1

        :param func: The callback given the logged-in user.
        :return: None.
        """

        class OnLogInCallback:
            """The callback when the user logs in."""

            @staticmethod
            def __call__(user: Any) -> None:
                """Runs the callback when the user logs in.

                :param user: The logged-in user.
                :return: None.
                """
                func(user)

        self.__on_login = OnLogInCallback()

    def init_app(self, app: Flask) -> None:
        """Initializes the Flask application.  The DigestAuth instance will
        be stored in ``app.extensions["digest_auth"]``.

        :Example:

        ::

            app: flask = Flask(__name__)
            auth: DigestAuth = DigestAuth()
            auth.init_app(app)

        :param app: The Flask application.
        :return: None.
        """
        app.extensions["digest_auth"] = self
        if "DIGEST_AUTH_REALM" in app.config:
            self.realm = app.config["DIGEST_AUTH_REALM"]

        if hasattr(app, "login_manager"):
            self.__init_login_manager(app)

    def __init_login_manager(self, app: Flask) -> None:
        """Initializes the Flask-Login login manager.

        :param app: The Flask application.
        :return: None.
        """
        from flask_login import LoginManager, login_user
        login_manager: LoginManager = getattr(app, "login_manager")

        @login_manager.unauthorized_handler
        def unauthorized() -> None:
            """Handles when the user is unauthorized.

            :return: None.
            """
            state: AuthState = getattr(request, "_digest_auth_state") \
                if hasattr(request, "_digest_auth_state") \
                else AuthState()
            response: Response = Response()
            response.status = 401
            response.headers["WWW-Authenticate"] \
                = self.__make_response_header(state)
            abort(response)

        @login_manager.request_loader
        def load_user_from_request(req: Request) -> Optional[Any]:
            """Loads the user from the request header.

            :param req: The request.
            :return: The authenticated user, or None if the
                authentication fails
            """
            request._digest_auth_state = AuthState()
            authorization: Authorization = req.authorization
            try:
                if authorization is None:
                    raise UnauthorizedException
                if authorization.type != "digest":
                    raise UnauthorizedException(
                        "Not an HTTP digest authorization")
                self.__authenticate(request._digest_auth_state)
                user = login_manager.user_callback(authorization.username)
                login_user(user)
                self.__on_login(user)
                return user
            except UnauthorizedException as e:
                if str(e) != "":
                    app.logger.warning(str(e))
                return None

    def logout(self) -> None:
        """Logs out the user.
        This actually causes the next authentication to fail, which forces
        the browser to ask the user for the username and password again.

        :Example:

        ::

            @app.post("/logout")
            @auth.login_required
            def logout():
                auth.logout()
                return redirect(request.form.get("next"))

        :return: None.
        """
        if "user" in session:
            del session["user"]
        try:
            if hasattr(current_app, "login_manager"):
                from flask_login import logout_user
                logout_user()
        except ModuleNotFoundError:
            pass
        session["digest_auth_logout"] = True


class AuthState:
    """The authentication state.  It keeps the status in the earlier
    authentication stage, so that the latter response stage knows how to
    response.
    """

    def __init__(self):
        """Constructs the authorization state."""
        self.opaque: Optional[str] = None
        """The opaque value specified by the client, if valid."""
        self.stale: Optional[bool] = None
        """The stale value, if there is a previous log in attempt."""


class UnauthorizedException(Exception):
    """The exception thrown when the authentication fails."""


class BasePasswordHashGetter:
    """The base callback that given the username, returns the password hash,
    or None if the user does not exist.  The default is to raise an
    :class:`UnboundLocalError` if the callback is not registered yet.

    See :meth:`flask_digest_auth.auth.DigestAuth.register_get_password`
    """

    @staticmethod
    def __call__(username: str) -> Optional[str]:
        """Returns the password hash of a user.

        :param username: The username.
        :return: The password hash, or None if the user does not exist.
        :raise UnboundLocalError: When the password hash getter function is
            not registered yet.
        """
        raise UnboundLocalError("The function to return the password hash"
                                " was not registered yet.")


class BaseUserGetter:
    """The base callback that given the username, returns the user, or None if
    the user does not exist.  The default is to raise an
    :class:`UnboundLocalError` if the callback is not registered yet.

    See :meth:`flask_digest_auth.auth.DigestAuth.register_get_user`
    """

    @staticmethod
    def __call__(username: str) -> Optional[Any]:
        """Returns a user.

        :param username: The username.
        :return: The user, or None if the user does not exist.
        :raise UnboundLocalError: When the user getter function is not
            registered yet.
        """
        raise UnboundLocalError("The function to return the user"
                                " was not registered yet.")


class BaseOnLogInCallback:
    """The base callback to run when the user logs in, given the logged-in
    user.  The default does nothing.

    See :meth:`flask_digest_auth.auth.DigestAuth.register_on_login`
    """

    @staticmethod
    def __call__(user: Any) -> None:
        """Runs the callback when the user logs in.

        :param user: The logged-in user.
        :return: None.
        """



# The Flask HTTP Digest Authentication Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2022/10/30

#  Copyright (c) 2022 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The test case for the HTTP digest authentication algorithm.

"""
import unittest
from typing import Optional, Literal

from flask_digest_auth import make_password_hash, calc_response


class AlgorithmTestCase(unittest.TestCase):
    """The test case for the HTTP digest authentication algorithm."""

    def test_response_value(self) -> None:
        """Tests the response value.
        See https://en.wikipedia.org/wiki/Digest_access_authentication.

        :return: None.
        """
        realm: str = "testrealm@host.com"
        username: str = "Mufasa"
        password: str = "Circle Of Life"
        method: str = "GET"
        uri: str = "/dir/index.html"
        nonce: str = "dcd98b7102dd2f0e8b11d0f600bfb0c093"
        qop: Optional[Literal["auth", "auth-int"]] = "auth"
        algorithm: Optional[Literal["MD5", "MD5-sess"]] = None
        cnonce: Optional[str] = "0a4f113b"
        nc: Optional[str] = "00000001"
        body: Optional[bytes] = None

        password_hash: str = make_password_hash(realm, username, password)
        response: str = calc_response(method, uri, password_hash, nonce, qop,
                                      algorithm, cnonce, nc, body)
        self.assertEqual(response, "6629fae49393a05397450978507c4ef1")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import sys

sys.path.insert(0, os.path.abspath('../../src/'))
import flask_digest_auth

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Flask-DigestAuth'
copyright = '2022-2023, imacat'
author = 'imacat'
release = flask_digest_auth.VERSION

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc"]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



# The Flask HTTP Digest Authentication Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2022/11/6

#  Copyright (c) 2022-2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The HTTP digest authentication.

"""
from flask_digest_auth.algo import make_password_hash, calc_response
from flask_digest_auth.auth import DigestAuth

VERSION: str = "0.7.1"
"""The package version."""



# The Flask HTTP Digest Authentication Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/10/5

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The common test libraries.

"""
from secrets import token_urlsafe
from typing import Optional, Literal, Dict

from werkzeug.datastructures import Authorization, WWWAuthenticate
from werkzeug.http import parse_set_header

from flask_digest_auth import calc_response, make_password_hash

REALM: str = "testrealm@host.com"
"""The realm."""
USERNAME: str = "Mufasa"
"""The username."""
PASSWORD: str = "Circle Of Life"
"""The password."""
ADMIN_1_URI: str = "/admin-1/auth"
"""The first administration URI."""
ADMIN_2_URI: str = "/admin-2/auth"
"""The first administration URI."""
LOGOUT_URI: str = "/logout"
"""The log out URI."""


def make_authorization(www_authenticate: WWWAuthenticate, uri: str,
                       username: str, password: str) -> str:
    """Composes and returns the request authorization.

    :param www_authenticate: The ``WWW-Authenticate`` response.
    :param uri: The request URI.
    :param username: The username.
    :param password: The password.
    :return: The request authorization header.
    """
    qop: Optional[Literal["auth", "auth-int"]] = None
    if "auth" in parse_set_header(www_authenticate.get("qop")):
        qop = "auth"

    cnonce: Optional[str] = None
    if qop is not None or www_authenticate.algorithm == "MD5-sess":
        cnonce = token_urlsafe(8)
    nc: Optional[str] = None
    count: int = 1
    if qop is not None:
        nc: str = hex(count)[2:].zfill(8)

    expected: str = calc_response(
        method="GET", uri=uri,
        password_hash=make_password_hash(www_authenticate.realm,
                                         username, password),
        nonce=www_authenticate.nonce, qop=qop,
        algorithm=www_authenticate.algorithm, cnonce=cnonce, nc=nc,
        body=None)

    data: Dict[str, str] = {
        "username": username, "realm": www_authenticate.realm,
        "nonce": www_authenticate.nonce, "uri": uri, "response": expected}
    if www_authenticate.algorithm is not None:
        data["algorithm"] = www_authenticate.algorithm
    if cnonce is not None:
        data["cnonce"] = cnonce
    if www_authenticate.opaque is not None:
        data["opaque"] = www_authenticate.opaque
    if qop is not None:
        data["qop"] = qop
    if nc is not None:
        data["nc"] = nc

    return str(Authorization("digest", data=data))
